﻿# -*- coding: utf-8 -*-

"""
This class implements the relations between python objects as a dictionary
with a list of objects.


"""
# python
from types import StringTypes

from BTrees.OOBTree import OOBTree
from BTrees.OOBTree import OOSet


class OneWayRelations:

    """
    Manages 'one way', 'one to many' relationships
    """

    def __init__(self):
        self.relations = OOBTree()


    def _toList(self, arg):
        """
        Make the argument into a list if it isn't allready
        >>> owr = OneWayRelations()
        >>> owr._toList('arg')
        ['arg']
        >>> owr._toList(u'arg')
        [u'arg']
        >>> owr._toList(['arg'])
        ['arg']
        >>> owr._toList(('arg',))
        ['arg']
        """
        if isinstance(arg, StringTypes):
            return [arg]
        return list(arg)


    def relate(self, keys, values):
        """
        Sets relations. If there allready is a key, it appends the 
        relations else it creates a new key with a list of relations
        
            >>> owr = OneWayRelations()
            >>> owr.relate(['maxm','mra'],['Group 1', 'Group 2', 'Group 3', ])
            >>> sorted(owr.relations['mra'])
            ['Group 1', 'Group 2', 'Group 3']
            
            >>> sorted(owr.relations['maxm'])
            ['Group 1', 'Group 2', 'Group 3']
            
            >>> owr.relations['Group 1']
            Traceback (most recent call last):
                ...
            KeyError: 'Group 1'
            
            >>> owr.relate(['tester','mra'],['Group 3', 'Group 4', 'Group 5', ])
            >>> owr.relate(['maxm','tester'],['Group 5', 'Group 6', ])
            >>> sorted(owr.relations['maxm'])
            ['Group 1', 'Group 2', 'Group 3', 'Group 5', 'Group 6']
        """
        keys = self._toList(keys)
        values = self._toList(values)
        # make shure there is something to relate
        if keys and values:
            relations = self.relations
            for key in keys:
                if not relations.has_key(key):
                    relations[key] = OOSet()
                relations[key].update(OOSet(values))


    def unrelate(self, keys, values):
        """
        Removes relations between objects, asserst all keys and values exists
        
        >>> owr = OneWayRelations()
        >>> owr.relate(['maxm','mra'],['Group 1', 'Group 2'])
        >>> owr.relate(['tester','mra'],['Group 3', 'Group 4'])
        >>> owr.relate(['maxm','tester'],['Group 5', 'Group 6', ])
        >>> sorted(owr.relations['maxm'])
        ['Group 1', 'Group 2', 'Group 5', 'Group 6']
        
        >>> owr.unrelate(['maxm'],['Group 5'])
        >>> sorted(owr.relations['maxm'])
        ['Group 1', 'Group 2', 'Group 6']
        
        >>> sorted(owr.relations['tester'])
        ['Group 3', 'Group 4', 'Group 5', 'Group 6']

        >>> owr.unrelate(['tester'],['Group 3', 'Group 4'])
        >>> sorted(owr.relations['tester'])
        ['Group 5', 'Group 6']
        
        >>> owr.unrelate(['tester'],['Group 5', 'Group 6'])
        >>> sorted(owr.relations['tester'])
        Traceback (most recent call last):
          ...
        KeyError: 'tester'
        """
        keys = self._toList(keys)
        values = self._toList(values)
        if keys and values:
            relations = self.relations
            for key in keys:
                oldVals = relations[key]
                for value in values:
                    if oldVals.has_key(value):
                        oldVals.remove(value)
                if len(oldVals) == 0:
                    del(self.relations[key])


    def delete(self, keys):
        """
        Removes all references to the relation/object.
        
        >>> owr = OneWayRelations()
        >>> owr.relate(['maxm','mra', 'tester'],['Group 1', 'Group 2', 'Group 3', ])
        >>> sorted(list(owr.relations.keys()))
        ['maxm', 'mra', 'tester']
        
        >>> owr.delete(['mra'])
        >>> sorted(list(owr.relations.keys()))
        ['maxm', 'tester']
        
        >>> owr.delete('tester')
        >>> sorted(list(owr.relations.keys()))
        ['maxm']
        
        >>> sorted(owr.relations['maxm'])
        ['Group 1', 'Group 2', 'Group 3']
        
        >>> owr.delete('non_existing_key')
        
        """
        keys = self._toList(keys)
        for key in keys:
            if self.relations.has_key(key):
                del(self.relations[key])


    def get(self, keys):
        """
        Returns all relations to the key, or an empty list
        
        >>> owr = OneWayRelations()
        >>> owr.relate(['maxm','mra'],['Group 1', 'Group 2', 'Group 3', ])
        >>> owr.relate(['tester', 'maxm'],['Group 4', 'Group 5', 'Group 6', ])
        
        >>> sorted(owr.get('mra'))
        ['Group 1', 'Group 2', 'Group 3']
        
        >>> sorted(owr.get(['mra', 'tester']))
        ['Group 1', 'Group 2', 'Group 3', 'Group 4', 'Group 5', 'Group 6']
        
        >>> sorted(owr.get(['maxm', 'tester']))
        ['Group 1', 'Group 2', 'Group 3', 'Group 4', 'Group 5', 'Group 6']
        
        >>> sorted(owr.get(['tester']))
        ['Group 4', 'Group 5', 'Group 6']
        
        >>> owr.get('non_existing_key')
        []
        """
        keys = self._toList(keys)
        uniq = {}
        for key in keys:
            for val in self.relations.get(key, []):
                uniq[val] = None
        return uniq.keys()





class WorkgroupRelations:

    """
    Implements 'two way', 'many to many' relations.
    """

    def __init__(self):
        # Store workgroups with workgroup UID as key and memberIds as value
        self._workgroups = OneWayRelations()
        # Store members with memberId as key and workgroup Uids as value
        self._members = OneWayRelations()


    def _addMembers(self, uids, memberIds):
        """
        Adds members to workgroups
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers(['Group 1', 'Group 2'], ['maxm','mra'])
        >>> sorted(wgr._getMembers('Group 1'))
        ['maxm', 'mra']
        
        >>> sorted(wgr._getMembers('Group not there'))
        []
        
        >>> sorted(wgr._getWorkgroups('maxm'))
        ['Group 1', 'Group 2']
        
        >>> wgr._addMembers(['Group 2', 'Group 3', 'Group 4'], ['mra', 'tester'])
        >>> sorted(wgr._getMembers('Group 3'))
        ['mra', 'tester']
        >>> sorted(wgr._getMembers('Group 2'))
        ['maxm', 'mra', 'tester']
       
        >>> sorted(wgr._getWorkgroups('mra'))
        ['Group 1', 'Group 2', 'Group 3', 'Group 4']
        
        >>> wgr._addMembers(['Group 5', 'Group 6'], ['maxm', 'tester'])
        >>> sorted(wgr._getMembers('Group 5'))
        ['maxm', 'tester']
        
        """
        self._workgroups.relate(uids, memberIds)
        self._members.relate(memberIds, uids)


    def _addWorkgroups(self, memberIds, uids):
        """
        Adds workgroup to a member

        >>> wgr = WorkgroupRelations()
        >>> wgr._addWorkgroups(['maxm','mra'], ['Group 1', 'Group 2'])
        >>> sorted(wgr._getMembers('Group 1'))
        ['maxm', 'mra']
        
        >>> sorted(wgr._getWorkgroups('Member not there'))
        []
        
        >>> sorted(wgr._getWorkgroups('maxm'))
        ['Group 1', 'Group 2']
        
        >>> wgr._addWorkgroups(['mra', 'tester'], ['Group 2', 'Group 3', 'Group 4'])
        >>> sorted(wgr._getMembers('Group 3'))
        ['mra', 'tester']
        >>> sorted(wgr._getMembers('Group 2'))
        ['maxm', 'mra', 'tester']
       
        >>> sorted(wgr._getWorkgroups('mra'))
        ['Group 1', 'Group 2', 'Group 3', 'Group 4']
        
        >>> wgr._addWorkgroups(['maxm', 'tester'], ['Group 5', 'Group 6'])
        >>> sorted(wgr._getMembers('Group 5'))
        ['maxm', 'tester']

        """
        self._members.relate(memberIds, uids)
        self._workgroups.relate(uids, memberIds)


    def _getMembers(self, uids):
        """
        Returns members of workgroup with uid
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers('Group 1', 'maxm')
        >>> wgr._getMembers('Group 1')
        ['maxm']
        
        >>> wgr._addMembers('Group 2', 'mra')
        >>> sorted(wgr._getMembers(['Group 1', 'Group 2']))
        ['maxm', 'mra']
        
        >>> wgr._getMembers('Group not there')
        []
        """
        return self._workgroups.get(uids)

    def _getAllMembers(self):
        """
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers('Group 1', 'maxm')
        >>> wgr._addMembers('Group 2', 'mra')
        >>> sorted(wgr._getAllMembers())
        ['maxm', 'mra']
        """
        return list(self._members.relations.keys())


    def _getWorkgroups(self, memberId):
        """
        Returns workgroups for member with memberId
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addWorkgroups('maxm', 'Group 1')
        >>> wgr._getWorkgroups('maxm')
        ['Group 1']
        
        >>> wgr._getMembers('Group not there')
        []
        """
        return self._members.get(memberId)


    def _getAllWorkgroups(self):
        """
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers('Group 1', 'maxm')
        >>> wgr._addMembers('Group 2', 'mra')
        >>> sorted(wgr._getAllWorkgroups())
        ['Group 1', 'Group 2']
        """
        return list(self._workgroups.relations.keys())


    def _unrelateMembers(self, uids, memberIds):
        """
        unrelates members from workgroups
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers('Group 1', ['maxm', 'tester'])
        >>> sorted(wgr._getMembers('Group 1'))
        ['maxm', 'tester']
        
        >>> wgr._unrelateMembers('Group 1', 'tester')
        >>> wgr._getMembers('Group 1')
        ['maxm']
        
        >>> wgr._unrelateMembers('Group 1', 'maxm')
        >>> wgr._getMembers('Group 1')
        []
        
        >>> wgr._getMembers('Group not there')
        []
        
        >>> keys = sorted(wgr._members.relations.keys())
        >>> keys
        []
        
        >>> keys = sorted(wgr._workgroups.relations.keys())
        >>> keys
        []
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers(['Group 1', 'Group 2', 'Group 3', ], ['maxm','mra'])
        >>> wgr._addMembers(['Group 3', 'Group 4', 'Group 5', ], ['tester','mra'])
        >>> wgr._addMembers(['Group 5', 'Group 6', ], ['maxm','tester'])
        >>> wgr._unrelateMembers('Group 1', 'mra')
        >>> wgr._getMembers('Group 1')
        ['maxm']
        >>> sorted(wgr._getMembers('Group 5'))
        ['maxm', 'mra', 'tester']
        >>> sorted(wgr._getWorkgroups('maxm'))
        ['Group 1', 'Group 2', 'Group 3', 'Group 5', 'Group 6']
        """
        self._workgroups.unrelate(uids, memberIds)
        self._members.unrelate(memberIds, uids)


    def _unrelateWorkgroups(self, memberIds, uids):
        """
        
        Unrelates workgroups from members
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addWorkgroups('maxm', ['Group 1', 'Group 2'])
        >>> sorted(wgr._getWorkgroups('maxm'))
        ['Group 1', 'Group 2']
        
        >>> wgr._unrelateWorkgroups('maxm', 'Group 1')
        >>> wgr._getWorkgroups('maxm')
        ['Group 2']
        
        >>> wgr._unrelateWorkgroups('maxm', 'Group 2', )
        >>> wgr._getWorkgroups('maxm')
        []
        
        >>> wgr._getMembers('Group 1')
        []
        
        >>> wgr._getWorkgroups('User not there')
        []
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers(['Group 1', 'Group 2', 'Group 3', ], ['maxm','mra'])
        >>> wgr._addMembers(['Group 3', 'Group 4', 'Group 5', ], ['tester','mra'])
        >>> wgr._addMembers(['Group 5', 'Group 6', ], ['maxm','tester'])
        >>> wgr._unrelateWorkgroups('mra', 'Group 1')
        >>> wgr._getMembers('Group 1')
        ['maxm']
        >>> sorted(wgr._getMembers('Group 5'))
        ['maxm', 'mra', 'tester']
        >>> sorted(wgr._getWorkgroups('maxm'))
        ['Group 1', 'Group 2', 'Group 3', 'Group 5', 'Group 6']
        >>> wgr._unrelateWorkgroups('maxm', 'Group 5')
        >>> sorted(wgr._getWorkgroups('maxm'))
        ['Group 1', 'Group 2', 'Group 3', 'Group 6']
        
        """
        self._members.unrelate(memberIds, uids)
        self._workgroups.unrelate(uids, memberIds)


    def _delMembers(self, memberIds):
        """
        Deletes members
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers(['Group 1', 'Group 2', 'Group 3', ], ['maxm','mra'])
        >>> wgr._addMembers(['Group 3', 'Group 4', 'Group 5', ], ['tester','mra'])
        >>> wgr._addMembers(['Group 5', 'Group 6', ], ['maxm','tester'])
        
        >>> wgr._getWorkgroups('maxm')
        ['Group 6', 'Group 5', 'Group 2', 'Group 3', 'Group 1']
        
        >>> wgr._delMembers('maxm')
        >>> wgr._getWorkgroups('maxm')
        []
        >>> wgr._getMembers('Group 1')
        ['mra']
        """
        self._unrelateMembers(self._getWorkgroups(memberIds), memberIds)


    def _delWorkgroups(self, uids):
        """
        Deletes workgroups
        
        >>> wgr = WorkgroupRelations()
        >>> wgr._addMembers(['Group 1', 'Group 2', 'Group 3', ], ['maxm','mra'])
        >>> wgr._addMembers(['Group 3', 'Group 4', 'Group 5', ], ['tester','mra'])
        >>> wgr._addMembers(['Group 5', 'Group 6', ], ['maxm','tester'])
        
        >>> sorted(wgr._getMembers('Group 1'))
        ['maxm', 'mra']
        
        >>> wgr._delWorkgroups('Group 1')
        >>> wgr._getMembers('Group 1')
        []
        
        >>> sorted(wgr._getWorkgroups('maxm'))
        ['Group 2', 'Group 3', 'Group 5', 'Group 6']

        >>> wgr._delWorkgroups(['Group 4', 'Group 5'])
        
        >>> sorted(wgr._getWorkgroups('maxm'))
        ['Group 2', 'Group 3', 'Group 6']
        
        >>> sorted(wgr._workgroups.relations.keys())
        ['Group 2', 'Group 3', 'Group 6']
        >>> sorted(wgr._workgroups.relations['Group 2'])
        ['maxm', 'mra']
        >>> sorted(wgr._workgroups.relations['Group 3'])
        ['maxm', 'mra', 'tester']
        >>> sorted(wgr._workgroups.relations['Group 6'])
        ['maxm', 'tester']
        
        >>> sorted(wgr._members.relations.keys())
        ['maxm', 'mra', 'tester']
        >>> sorted(wgr._members.relations['maxm'])
        ['Group 2', 'Group 3', 'Group 6']
        
        >>> sorted([sorted(s) for s in wgr._members.relations.values()])
        [['Group 2', 'Group 3'], ['Group 2', 'Group 3', 'Group 6'], ['Group 3', 'Group 6']]
        
        """
        self._unrelateWorkgroups(self._getMembers(uids), uids)














def sorted(ls):
    "Doctest helper"
    ls = list(ls)
    ls.sort()
    return ls

if __name__ == "__main__":
    
    import os.path, doctest, WorkgroupRelations
    # import and test this file
    doctest.testmod(WorkgroupRelations)
